unit ChipThr;

interface
uses Windows,Classes,Chip;

 type
  TBasicClass=class of TBasic;
  TSendKeyEvent=procedure (Sender:TObject; Key:char) of object;
  TBasicThread=class(TThread)
  protected
   FTag:integer;
   FOnIdle:TNotifyEvent;
   FOnSendKey:TSendKeyEvent;
   FKeyBuf:string;
   FKeyLock:TRTLCriticalSection;
   FCommand:string;
   FBas:TBasic;
   FOut:char;
   procedure DoTTY;
  public
   function ReadKey:char;
   function KeyPressed:boolean;
  protected
   procedure TTY(c:char);
   function ReadBuf(Buffer: PChar; Count: Cardinal): Cardinal;
   procedure WriteBuf(Buffer: PChar; Count: Cardinal);
   procedure Execute;override;
   procedure Idle;
   procedure InitializeEngine;virtual;
  public
   constructor Create(AFactory:TBasicClass; const ACommand:string);
   destructor Destroy;override;
   procedure PushKey(key:char);
   property OnSendKey:TSendKeyEvent read FOnSendKey write FOnSendKey;
   property OnIdle:TNotifyEvent read FOnIdle write FOnIdle;
   property Tag:integer read FTag write FTag;
   property Script:TBasic read FBas;
  end;

implementation
uses SysUtils;

 procedure TBasicThread.InitializeEngine;
 begin
 end;

 procedure TBasicThread.Idle;
 begin
  if Assigned(FOnIdle) then FOnIdle(Self);
  Suspend
 end;

 procedure TBasicThread.DoTTY;
 begin
  if Assigned(FOnSendKey) then FOnSendKey(Self,FOut)
 end;

 procedure TBasicThread.TTY;
 begin
  FOut:=c;
  Synchronize(DoTTY)
 end;

 constructor TBasicThread.Create;
 begin
  inherited Create(true);
  FCommand:=ACommand;
  FBas:=AFactory.Create;
  InitializeCriticalSection(FKeyLock)
 end;

 function TBasicThread.KeyPressed;
 begin
  EnterCriticalSection(FKeyLock);
  Result:=FKeyBuf<>'';
  LeaveCriticalSection(FKeyLock)
 end;

 function TBasicThread.ReadKey;
 label loop;
 begin
  loop:
  EnterCriticalSection(FKeyLock);
  if FKeyBuf='' then
   Result:=#0
  else begin
   Result:=FKeyBuf[1];
   Delete(FKeyBuf,1,1)
  end;
  LeaveCriticalSection(FKeyLock);
  if Result=#0 then begin
   Idle;
   goto loop
  end
 end;

 function TBasicThread.ReadBuf;
 var
   Ch: Char;
   I: Cardinal;

   procedure WriteChar(c:char);
   begin
    TTY(c)
   end;

 begin
   I := 0;
   repeat
     Ch := ReadKey;
     case Ch of
       #8:
         if I > 0 then
         begin
           Dec(I);
           WriteChar(#8);
         end;
       #32..#255:
         if I < Count - 2 then
         begin
           Buffer[I] := Ch;
           Inc(I);
           WriteChar(Ch);
         end;
     end;
   until (Ch in [#0,#13]);
   Buffer[I] := Ch;
   Inc(I);
   if Ch = #13 then
   begin
     Buffer[I] := #10;
     Inc(I);
     WriteBuf(#13#10,2);
   end;
   ReadBuf := I;
 end;

 procedure TBasicThread.WriteBuf;
 begin
  while Count>0 do begin
   TTY(Buffer^);
   inc(Buffer);
   dec(Count)
  end
 end;

 { Text file device driver output function }

 function CrtOutput(var F: TTextRec): Integer; far;
 begin
   if F.BufPos <> 0 then
   with TObject((@F.UserData)^) as TBasicThread do
   begin
     WriteBuf(PChar(F.BufPtr), F.BufPos);
     F.BufPos := 0;
   end;
   CrtOutput := 0;
 end;

 { Text file device driver input function }

 function CrtInput(var F: TTextRec): Integer; far;
 begin
   with TObject((@F.UserData)^) as TBasicThread do
     F.BufEnd := ReadBuf(PChar(F.BufPtr), F.BufSize);
   F.BufPos := 0;
   CrtInput := 0;
 end;

 { Text file device driver close function }

 function CrtClose(var F: TTextRec): Integer; far;
 begin
   CrtClose := 0;
 end;

 { Text file device driver open function }

 function CrtOpen(var F: TTextRec): Integer; far;
 begin
   if F.Mode = fmInput then
   begin
     F.InOutFunc := @CrtInput;
     F.FlushFunc := nil;
   end else
   begin
     F.Mode := fmOutput;
     F.InOutFunc := @CrtOutput;
     F.FlushFunc := @CrtOutput;
   end;
   F.CloseFunc := @CrtClose;
   CrtOpen := 0;
 end;

 procedure TBasicThread.Execute;

  procedure AssignCrt(var F: Text);
  begin
    with TTextRec(F) do
    begin
      Handle := Cardinal(-1);
      Mode := fmClosed;
      BufSize := SizeOf(Buffer);
      BufPtr := @Buffer;
      OpenFunc := @CrtOpen;
      Move(Self, UserData[1],Sizeof(Pointer));
      Name[0] := #0;
    end;
  end;

 begin
  AssignCrt(FBas.Input);
  Reset(FBas.Input);
  AssignCrt(FBas.Output);
  Rewrite(FBas.Output);
  InitializeEngine;
  FBas.Run(FCommand)
 end;

 procedure TBasicThread.PushKey;
 begin
  EnterCriticalSection(FKeyLock);
  FKeyBuf:=FKeyBuf+key;
  Resume;
  LeaveCriticalSection(FKeyLock)
 end;

 destructor TBasicThread.Destroy;
 begin
  FBas.Free;
  DeleteCriticalSection(FKeyLock);
  inherited
 end;

end.
